'use strict';

const argon2 = require('argon2');
const crypto = require('crypto');
const common = require('./common');
const Enum = require('./enum');
const { ResponseError } = require('./errors');

const _fileScope = common.fileScope(__filename);

/**
 * Let us roll our own authentication dingus.
 * For 'admin' level access, we allow this custom header scheme, or optionally Basic if a password is stored.
 * We shall authenticate an id/secret from a header of this format:
 *   Signature: id:epoch:hmac(id + epoch + body)
 * 
 * For 'normal' level access, we allow a token passed as a parameter or a Bearer token.
 */

const defaultOptions = {
  basicRealm: 'Authenticate',
  customAuthName: 'SqueepSignature',
  timestampGraceMs: 120 * 1000, // In ms
};

class Authenticator {
  constructor(logger, db, options = {}) {
    common.setOptions(this, defaultOptions, options);
    this.logger = logger;
    this.db = db;
  }


  /**
   * For Reasons, get rid of iffy b64 characters.
   * @param {string} s 
   */
  static b64Armor(s) {
    return s && s.replace(/\+/g, '-').replace(/\//g, '_').replace(/=/g, '.');
  }


  /**
   * Put them back, if we need to.
   * @param {string} s 
   */
  static b64Unarmor(s) {
    return s && s.replace(/-/g, '+').replace(/_/g, '/').replace(/\./g, '=');
  }


  /**
   * 
   * @param {string} secret 
   * @param {string} id 
   * @param {string} epoch 
   * @param {string} body 
   */
  static signature(secret, id, epoch, body) {
    const hmac = crypto.createHmac('sha384', secret);
    hmac.update(new Buffer.from(id));
    hmac.update(new Buffer.from(epoch.toString()));
    hmac.update(new Buffer.from(body));
    const digest = hmac.digest('base64');
    return Authenticator.b64Armor(digest);
  }


  /**
   * 
   * @param {number} size 
   */
  static async generateToken(size = 36) {
    const randomBuffer = await common.randomBytesAsync(size);
    const token = randomBuffer.toString('base64');
    return Authenticator.b64Armor(token);
  }


  /**
   * 
   * @param {http.ServerResponse} res 
   */
  requestSignature(res) {
    res.setHeader(Enum.Header.WWWAuthenticate, `${this.customAuthName}`);
    throw new ResponseError(Enum.ErrorResponse.Unauthorized);
  }


  /**
   * Check our bespoke scheme.  If valid, update context with auth details.
   * @param {String} reqSig 
   * @param {Object} ctx 
   * @param {string} ctx.rawBody
   */
  async isValidSignature(reqSig, ctx) {
    const _scope = _fileScope('isValidSignature');

    this.logger.debug(_scope, 'checking authentication', { reqSig });

    const reqSigParts = (reqSig || '').split(':');

    if (reqSigParts.length !== 3) {
      this.logger.debug(_scope, 'unparsable', { reqSigParts });
      return false;
    }

    // Get secret for id or fail.
    const authenticationId = reqSigParts[0];
    let authData;
    await this.db.context(async (dbCtx) => {
      authData = await this.db.getAuthById(dbCtx, authenticationId);
    });
    const secret = authData && authData.secret;

    if (!secret) {
      this.logger.debug(_scope, 'failed, invalid authenticationId', { ctx })
      return false;
    }

    // Check timestamp
    ctx.authenticationEpoch = parseInt(reqSigParts[1], 10);
    const currentEpoch = Date.now() / 1000;
    ctx.authenticationDrift = currentEpoch - ctx.authenticationEpoch;
    if (Math.abs(ctx.authenticationDrift) > this.timestampGraceMs) {
      this.logger.debug(_scope, 'failed, invalid timestamp', { ctx })
      return false;
    }

    // Check signature
    ctx.authenticationSignatureProvided = reqSigParts[2];
    ctx.authenticationSignatureComputed = Authenticator.signature(secret, authenticationId, ctx.authenticationEpoch.toString(), ctx.rawBody);
    if (ctx.authenticationSignatureProvided !== ctx.authenticationSignatureComputed) {
      this.logger.debug(_scope, 'failed, signature mismatch', { ctx });
      return false;
    }
    
    ctx.authenticationId = authenticationId;
    ctx.authenticationSecret = secret; // Retain this for computing response signature

    return true; // Success
  }


  /**
   * Check for valid basic auth.
   * @param {string} credentials 
   * @param {object} ctx 
   */
  async isValidBasic(credentials, ctx) {
    const _scope = _fileScope('isValidBasic');
    this.logger.debug(_scope, 'called', { ctx });

    const [ authenticationId, authenticationPass ] = common.splitFirst(credentials, ':', '');

    let authData;
    await this.db.context(async (dbCtx) => {
      authData = await this.db.getAuthById(dbCtx, authenticationId);
    });
    if (!authData) {
      this.logger.debug(_scope, 'failed, invalid authenticationId', { ctx });
      return false;
    }

    // Update pwhash
    // authData.password = await argon2.hash(newPassword, { type: argon2.id });
    if (authData.password.startsWith('$argon2')) {
      if (await argon2.verify(authdata.password, authenticationPass)) {
        this.logger.debug(_scope, 'failed argon2 verify', { ctx });
        return false;
      } else {
        this.logger.debug(_scope, 'passed argon2 verify', { ctx });
      }
    } else {
      if (authData.password !== authenticationPass) {
        this.logger.debug(_scope, 'failed, password mismatch', { ctx });
        return false;
      }
    }

    ctx.authenticationId = authenticationId;

    return true; // Success
  }


  /**
   * Check if a presented Bearer token matches the presented link.
   * @param {string} credentials 
   * @param {object} ctx 
   */
  async isValidBearer(credentials, ctx) {
    const _scope = _fileScope('isValidBearer');
    this.logger.debug(_scope, 'called', { ctx });

    const linkId = ctx.params.id;
    return await this.isValidToken(linkId, credentials);
  }


  /**
   * 
   * @param {string} linkId 
   * @param {string} token 
   */
  async isValidToken(linkId, token) {
    const _scope = _fileScope('isValidToken');
    this.logger.debug(_scope, 'called', { linkId, token });

    if (!token) {
      this.logger.debug(_scope, 'no token, ignoring', { linkId, token });
      return false;
    }

    if (!linkId) {
      this.logger.debug(_scope, 'no id, ignoring token', { linkId, token });
      return false;
    }

    let link;
    await this.db.context(async (dbCtx) => {
      link = await this.db.getLinkById(dbCtx, linkId);
    });

    if (!link) {
      this.logger.debug(_scope, 'failed, invalid id', { linkId, token });
      return false;
    }

    // eslint-disable-next-line security/detect-possible-timing-attacks
    if (link.authToken !== token) {
      this.logger.debug(_scope, 'failed, token mismatch', { linkId, token });
      return false;
    }

    // Success
    return true;
  }


  /**
   * Determine which sort of Authorization header is available, and either parse it or request one.
   * @param {string} authorizationHeader
   * @param {object} ctx 
   */
  async isValidAuthorization(authorizationHeader, ctx) {
    const _scope = _fileScope('isValidAuthorization');
    this.logger.debug(_scope, 'called', { authorizationHeader, ctx });

    const [ authMethod, b64Credentials ] = common.splitFirst(authorizationHeader, ' ', '').map((x) => x.trim());
    const credentials = Buffer.from(b64Credentials, 'base64').toString('utf-8');
    // eslint-disable-next-line sonarjs/no-small-switch
    switch (authMethod.toLowerCase()) {
      case 'basic':
        return await this.isValidBasic(credentials, ctx);

      case 'bearer':
        return await this.isValidBearer(credentials, ctx);

      default:
        return false;
    }
  }


  /**
   * 
   * @param {http.ServerResponse} res 
   */
  requestBasic(res) {
    res.setHeader(Enum.Header.WWWAuthenticate, `Basic realm="${this.basicRealm}", charset="UTF-8"`);
    throw new ResponseError(Enum.ErrorResponse.Unauthorized);
  }


  /**
   * Require a request passes auth.
   * Sets ctx.authenticationId on valid non-token authentication.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async required(req, res, ctx) {
    const _scope = _fileScope('required');
    this.logger.debug(_scope, 'called', { ctx });

    let authData;

    // Prefer custom
    authData = req.getHeader(Enum.Header.Signature);
    if (authData) {
      const validSignature = await this.isValidSignature(authData, ctx);
      if (validSignature) {
        return true; // Success
      }
      // Signature was present, but failed.  Request signature auth.
      return this.requestSignature(res);
    }

    // Accept standards
    authData = req.getHeader(Enum.Header.Authorization);
    if (authData) {
      const validAuthorization = await this.isValidAuthorization(authData, ctx);
      if (validAuthorization) {
        return true; // Success
      }
      return this.requestBasic(res);
    }

    const linkId = ctx.params && ctx.params.id;
    // If there is an id parameter, check for a valid token query parameter
    if (linkId) {
      authData = (ctx.queryParams && ctx.queryParams.token) || (ctx.parsedBody && ctx.parsedBody.token);
      if (authData) {
        const validToken = await this.isValidToken(linkId, authData);
        if (validToken) {
          return true; // Success
        }
      }
    }

    return this.requestBasic(res);
  }


  /**
   * Check and set authentication if provided and valid, but do not require it.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async optional(req, res, ctx) {
    const _scope = _fileScope('optional');
    this.logger.debug(_scope, 'called', { ctx });

    let authData;

    // Allow and prefer a valid signature.
    authData = req.getHeader(Enum.Header.Signature);
    if (authData) {
      const validSignature = await this.isValidSignature(authData, ctx);
      if (validSignature) {
        return true; // Success
      }
    }

    // Allow valid auth.
    authData = req.getHeader(Enum.Header.Authorization);
    if (authData) {
      const validAuthorization = await this.isValidAuthorization(authData, ctx);
      if (validAuthorization) {
        return true; // Success
      }
    }

    // Allow a valid plain token.
    const linkId = ctx.params && ctx.params.id;
    if (linkId) {
      const token = (ctx.queryParams && ctx.queryParams.token) || (ctx.parsedBody && ctx.parsedBody.token);
      if (token) {
        const validToken = await this.isValidToken(linkId, token);
        if (validToken) {
          return true; // Success
        }
      }
    }

    return false;
  }


  /**
   * If a request included signature auth, respond in kind.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  signResponse(req, res, ctx) {
    const _scope = _fileScope('signResponse');
    
    // Keep huge payloads out of logs...
    const logCtx = Object.assign({}, ctx);
    if (logCtx.responseBody) {
      if (typeof logCtx.responseBody !== 'string') {
        logCtx.responseBody = logCtx.responseBody.toString();
      }
      if (logCtx.responseBody.length > 128) {
        logCtx.responseBody = logCtx.responseBody.slice(0, 128) + `... (${logCtx.responseBody.length} bytes)`;
      }
    }
    this.logger.debug(_scope, 'called', { logCtx });

    if (ctx.authenticationSecret && ctx.authenticationId) {
      const currentEpoch = Date.now() / 1000;
      const signature = Authenticator.signature(ctx.authenticationSecret, ctx.authenticationId, currentEpoch, ctx.responseBody || '');
      const signatureHeader = [ctx.authenticationId, currentEpoch, signature].join(':');
      res.setHeader(Enum.Header.Signature, signatureHeader);
      this.logger.debug(_scope, 'added signature', { logCtx, signatureHeader });
    }
  }

}

module.exports = Authenticator;
