'use strict';

const pgpInitOptions = {
  capSQL: true,
};

const path = require('path');
const pgp = require('pg-promise')(pgpInitOptions);
const BaseDatabase = require('../base');
const common = require('../../common');
const DBErrors = require('../errors');

const _fileScope = common.fileScope(__filename);

const defaultOptions = {
  connectionString: undefined,
  queryLogLevel: undefined,
  listenerPingDelayMs: 2000,
  listenerRetryDelayMs: 5000,
  listenerRetryTimes: 10,
  listenerChannel: 'cache_invalidation',
  listenerCallback: () => {},
  listenerLostCallback: () => {},
};

class PostgresDatabase extends BaseDatabase {
  constructor(logger, options, _pgp = pgp) {
    const _scope = _fileScope('constructor');

    super(logger);
    common.setOptions(this, defaultOptions, options);

    this.logger.debug(_scope, 'connecting', { connectionString: this.connectionString });

    this.db = _pgp(this.connectionString);

    if (this.queryLogLevel) {
      pgpInitOptions.query = (e) => {
        this.logger[this.queryLogLevel](_fileScope('query'), e.query, { params: e.params });
      };
    }

    pgpInitOptions.error = (err, e) => {
      this.logger[this.queryLogLevel](_fileScope('pgp'), '', { err, e });
    };

    pgpInitOptions.receive = (data, result, e) => {
      const exemplaryRow = data[0];
      for (const prop in exemplaryRow) {
        const camel = BaseDatabase._camelfy(prop);
        if (!(camel in exemplaryRow)) {
          for (const d of data) {
            // eslint-disable-next-line security/detect-object-injection
            d[camel] = d[prop];
            // eslint-disable-next-line security/detect-object-injection
            delete d[prop];
          }
        }
      }

      if (this.queryLogLevel) {
        this.logger[this.queryLogLevel](_fileScope('result'), e.query, PostgresDatabase._resultLog(result));
      }
    };
    this._initStatements(_pgp);
  }

  static _resultLog(result) {
    return {
      command: result.commaand,
      rowCount: result.rowCount,
      duration: result.duration,
    };
  }

  _initStatements(_pgp) {
    const _scope = _fileScope('_initStatements');

    const qfOptions = {
      minify: true,
    };
    this.statement = _pgp.utils.enumSql(path.join(__dirname, 'sql'), {}, (file) => {
      const qf = new _pgp.QueryFile(file, qfOptions);
      this.logger.debug(_scope, file, { file: qf.file, options: qf.options, query: qf.query });
      if (qf.error) {
        this.logger.error(_scope, file, { error: qf.error });
        throw qf.error;
      }
      return qf;
    });

    this.logger.debug(_scope, 'complete', { statementCount: Object.keys(this.statement).length });
  }

  // WIP not used yet
  async _initListener() {
    await this._reconnectListener(0, 1);
    this._sendListenerNotifications();
  }
  _sendListenerNotifications() {
    const _scope = _fileScope('_sendListenerNotifications');
    setTimeout(async () => {
      if (this.listenerConnection) {
        try {
          await this.listenerConnection.none('NOTIFY $1~, $2', [this.listenerChannel, 'ping']);
        } catch (e) {
          this.logger.error(_scope, 'failed', e);
        } finally {
          this._sendListenerNotifications();
        }
      }
    }, this.listenerPingDelayMs);
  }
  async _onListenerConnectionLost(err, ev) {
    const _scope = _fileScope('_onConnectionLost');
    const eventName = 'notification';
    this.listenerConnection = null;
    this.logger.error(_scope, 'listener connection lost', { err, ev });
    ev.client.removeListener(eventName, this._onListenerNotificationBound);
    await this.listenerLostCallback();
    await this._reconnectListener(this.listenerRetryDelayMs, this.listenerRetryTimes);
    this.logger.debug(_scope, 'listener reconnected');
  }
  async _onListenerNotification(data) {
    return await this.listenerCallback(data.payload);
  }
  async _reconnectListener(delay, retriesRemaining) {
    const _scope = _fileScope('_reconnectListener');
    const eventName = 'notification';
    if (this.listenerConnection) {
      this.listenerConnection.done();
      this.listenerConnection = null;
    }
    if (this.listenerReconnectPending) {
      clearTimeout(this.listenerReconnectPending);
      delete this.listenerReconnectPending;
    }
    return new Promise((resolve, reject) => {
      this.listenerReconnectPending = setTimeout(async () => {
        try {
          this.listenerConnection = await this.db.connect({
            direct: true,
            onLost: this._onListenerConnectionLost.bind(this),
          });
          if (!this._onListenerNotificationBound) {
            this._onListenerNotificationBound = this._onListenerNotification.bind(this);
          }
          this.listenerConnection.client.on(eventName, this._onListenerNotificationBound);
          await this.listenerConnection.none('LISTEN $1~', this.listenerChannel);
          delete this.listenerReconnectPending;
          this.logger.debug(_scope, 'listener connection established');
          resolve();
        } catch (e) {
          if (retriesRemaining > 0) {
            try {
              await this._reconnectListener(delay, retriesRemaining - 1);
              resolve();
            } catch (e2) {
              reject(e2);
            }
          } else {
            reject(e);
          }
        }
      }, delay);
    });
  }


  // eslint-disable-next-line class-methods-use-this
  _postgresInfo(result) {
    return {
      changes: result.rowCount,
      lastInsertRowid: result.rows.length ? result.rows[0].id : undefined,
      duration: result.duration,
    };
  }

  async context(fn) {
    return await this.db.task(async (t) => await fn(t));
  }


  async transaction(dbCtx, fn) {
    dbCtx = dbCtx || this.db;
    return await dbCtx.txIf(async (t) => await fn(t));
  }


  async getAuthById(dbCtx, id) {
    const _scope = _fileScope('getAuthById');
    this.logger.debug(_scope, 'called', { id });
  
    dbCtx = dbCtx || this.db;

    const auth = await dbCtx.oneOrNone(this.statement.authGetById, { id });
    this.logger.debug(_scope, 'get', { auth });
    return auth;
  }


  static _epochFix(epoch) {
    switch (epoch) {
      case Infinity:
        return Number.MAX_SAFE_INTEGER;
      
      case -Infinity:
        return 0;

      default:
        return epoch;
    }
  }


  static _linkToNative(link) {
    return link && {
      ...link,
      created: PostgresDatabase._epochFix(link.created),
      lastAccess: PostgresDatabase._epochFix(link.lastAccess),
      expires: PostgresDatabase._epochFix(link.expires),
    };
  }

  async insertLink(dbCtx, id, url, authToken) {
    const _scope = _fileScope('insertLink');
    this.logger.debug(_scope, 'called', { id, url, authToken });
  
    dbCtx = dbCtx || this.db;

    const result = await dbCtx.result(this.statement.linkUpsert, { id, url, authToken });
    this.logger.debug(_scope, 'result', PostgresDatabase._resultLog(result) );
    return this._postgresInfo(result);
  }


  async getLinkById(dbCtx, id) {
    const _scope = _fileScope('getLinkById');
    this.logger.debug(_scope, 'called', { id });
  
    dbCtx = dbCtx || this.db;

    const link = await dbCtx.oneOrNone(this.statement.linkGetById, { id });
    this.logger.debug(_scope, 'get', { link });
    return PostgresDatabase._linkToNative(link);
  }


  async getLinkByUrl(dbCtx, url) {
    const _scope = _fileScope('getLinkByUrl');
    this.logger.debug(_scope, 'called', { url });
  
    dbCtx = dbCtx || this.db;

    const link = await dbCtx.oneOrNone(this.statement.linkGetByUrl, { url });
    this.logger.debug(_scope, 'get', { link });
    return PostgresDatabase._linkToNative(link);
  }


  async accessLink(dbCtx, id) {
    const _scope = _fileScope('accessLink');
    this.logger.debug(_scope, 'called', { id });
  
    dbCtx = dbCtx || this.db;

    const link = await dbCtx.oneOrNone(this.statement.linkAccess, { id });
    this.logger.debug(_scope, 'get', { link });
    return PostgresDatabase._linkToNative(link);
  }


  async expireLink(dbCtx, id, expires) {
    const _scope = _fileScope('expireLink');
    this.logger.debug(_scope, 'called', { id, expires });
  
    dbCtx = dbCtx || this.db;

    const result = await dbCtx.result(this.statement.linkExpire, { expires, id });
    this.logger.debug(_scope, 'result', PostgresDatabase._resultLog(result) );
    return this._postgresInfo(result);
  }


  async updateLink(dbCtx, id, url) {
    const _scope = _fileScope('updateLink');
    this.logger.debug(_scope, 'called', { id, url });
  
    dbCtx = dbCtx || this.db;

    const result = await dbCtx.result(this.statement.linkUpdate, { id, url });
    this.logger.debug(_scope, 'result', PostgresDatabase._resultLog(result) );
    return this._postgresInfo(result);
  }


  async getAllLinks(dbCtx) {
    const _scope = _fileScope('getAllLinks');
    this.logger.debug(_scope, 'called', { });
  
    dbCtx = dbCtx || this.db;

    const links = await dbCtx.manyOrNone(this.statement.linkGetAll, { });
    this.logger.debug(_scope, 'get', { links });
    return links.map((l) => PostgresDatabase._linkToNative(l));
  }


}

module.exports = PostgresDatabase;
