'use strict';

const SQLite = require('better-sqlite3');
// XXX: const NodeCleanup = require('node-cleanup');
const BaseDatabase = require('../base');
const common = require('../../common');
const DBErrors = require('../errors');

const _fileScope = common.fileScope(__filename);

const defaultOptions = {
  connectionString: '',
  optimizeAfterChanges: 1000,
  queryLogLevel: 'debug',
};

const EPOCH_NOW = '(strftime(\'%s\', \'now\'))';

class SQLiteDatabase extends BaseDatabase {
  constructor(logger, options) {
    const _scope = _fileScope('constructor');

    super(logger);
    common.setOptions(this, defaultOptions, options);

    this.dbFilename = this.connectionString.slice('sqlite://'.length) || ':memory:';

    this.logger.debug(_scope, 'connecting', { dbFilename: this.dbFilename });

    this.db = new SQLite(this.dbFilename, {
      verbose: (query) => this.queryLogLevel && this.logger[this.queryLogLevel](_fileScope('statement'), 'executing', { query }),
    });

    this.changesSinceLastOptimize = BigInt(0);

    /*
    NodeCleanup(() => {
      this._optimize();
      this._closeConnection();
    });
    */

    this.db.pragma('foreign_keys = on');
    this.db.pragma('journal_mode = WAL');
    this.db.defaultSafeIntegers(true);

    this._initTables();
    this._initStatements();
  }

  _initTables() {
    const _scope = _fileScope('_initTables');
    const tableExists = this.db.prepare('SELECT name FROM sqlite_master WHERE type=\'table\' AND name=:tableName').pluck(true);

    const tables = [
      {
        name: 'auth',
        statements: [
          `CREATE TABLE IF NOT EXISTS auth (
            id TEXT NOT NULL PRIMARY KEY,
            secret TEXT NOT NULL,
            password TEXT
          )`,
          `INSERT INTO auth (id, secret, password)
            VALUES
            ('foo', 'secret', 'quux')`,
        ],
      },
      {
        name: 'link',
        statements: [
          `CREATE TABLE IF NOT EXISTS link (
            id TEXT NOT NULL PRIMARY KEY,
            url TEXT NOT NULL UNIQUE,
            created INTEGER NOT NULL DEFAULT ${EPOCH_NOW},
            last_access INTEGER NOT NULL DEFAULT 0,
            accesses INTEGER NOT NULL DEFAULT 0,
            expires INTEGER,
            auth_token TEXT
          )`,
          'CREATE INDEX IF NOT EXISTS link_url_idx ON link(url)',
        ],
      },
    ];

    tables.forEach((t) => {
      const table = tableExists.get({ tableName: t.name });
      if (table === undefined) {
        t.statements.forEach((s) => {
          this.logger.info(_scope, 'creating table', { tableName: t.name });
          this.db.prepare(s).run();
        });
      }
    });
  }

  _initStatements() {
    this.statement = {
      getAuthById: this.db.prepare('SELECT * FROM auth WHERE id = :id'),
      getLinkById: this.db.prepare('SELECT * FROM link WHERE id = :id'),
      getLinkByUrl: this.db.prepare('SELECT * FROM link WHERE url = :url'),
      insertLink: this.db.prepare('INSERT INTO link (id, url, auth_token) VALUES (:id, :url, :authToken)'),
      updateLink: this.db.prepare('UPDATE link SET url = :url WHERE id = :id'),
      incrementLinkAccess: this.db.prepare(`UPDATE link SET accesses = accesses + 1, last_access = ${EPOCH_NOW} WHERE id = :id`),
      expireLink: this.db.prepare('UPDATE link SET expires = :expires WHERE id = :id'),
      linkGetAll: this.db.prepare('SELECT * FROM link'),
    };
  }

  _optimize() {
    const _scope = _fileScope('_optimize');
    this.logger.debug(_scope, 'called', {});
    const optimizations = this.db.prepare('SELECT * FROM pragma_optimize(0x03)').all();
    this.logger.debug(_scope, 'pragma preview', { optimizations });
    this.db.pragma('optimize');
    this.changesSinceLastOptimize = BigInt(0);
  }

  _maybeOptimize() {
    if (this.changesSinceLastOptimize >= this.optimizeAfterChanges) {
      this._optimize();
    }
  }

  _sqliteInfo(info) {
    if (info.changes) {
      this.changesSinceLastOptimize += BigInt(info.changes);
      this._maybeOptimize();
    }
    return {
      changes: Number(info.changes),
      lastInsertRowid: Number(info.lastInsertRowid),
    };
  }

  static _deOphidiate(row) {
    return row && Object.keys(row).reduce((snakelessRow, k) => Object.assign(snakelessRow, {
      // eslint-disable-next-line security/detect-object-injection
      [BaseDatabase._camelfy(k)]: row[k],
    }), {});
  }


  async context(fn) {
    const dbCtx = this.db;
    return await fn(dbCtx);
  }

  async transaction(dbCtx, fn, ...rest) {
    dbCtx = dbCtx || this.db;
    return await dbCtx.transaction(async (...args) => await fn(...args))(dbCtx, ...rest);
  }

  async getAuthById(dbCtx, id) {
    const _scope = _fileScope('getAuthById');
    this.logger.debug(_scope, 'called', { id });
  
    let auth;
    auth = this.statement.getAuthById.get({ id });
    auth = SQLiteDatabase._deOphidiate(auth);

    this.logger.debug(_scope, 'get', { auth });
    return auth;
  }

  async insertLink(dbCtx, id, url, authToken) {
    const _scope = _fileScope('insertLink');
    this.logger.debug(_scope, 'called', { id, url });

    let info;
    try {
      info = this.statement.insertLink.run({ id, url, authToken });
    } catch (e) {
      switch (e.code) {
        case 'SQLITE_CONSTRAINT_UNIQUE':
        case 'SQLITE_CONSTRAINT_PRIMARYKEY': {
          this.logger.debug(_scope, 'updating existing id', { id, url });
          info = this.statement.updateLink.run({ id, url });
          break;
        }

        default: {
          this.logger.error(_scope, 'failed to insert link', { error: e, id, url });
          throw e;
        }
      }
    }
    this.logger.debug(_scope, 'run', { info });
    if (info.changes != 1) {
      this.logger.error(_scope, 'failed to insert link', { id, url, info });
      throw new DBErrors.UnexpectedResult();
    }

    return this._sqliteInfo(info);
  }

  static _linkToNative(link) {
    return link && {
      id: link.id,
      url: link.url,
      created: Number(link.created),
      lastAccess: Number(link.lastAccess),
      accesses: Number(link.accesses),
      expires: ('expires' in link) ? Number(link.expires) : undefined,
      authToken: link.authToken,
    };
  }

  async getLinkById(dbCtx, id) {
    const _scope = _fileScope('getLinkById');
    this.logger.debug(_scope, 'called', { id });

    let link = this.statement.getLinkById.get({ id });
    link = SQLiteDatabase._deOphidiate(link);

    this.logger.debug(_scope, 'get', { link });
    return SQLiteDatabase._linkToNative(link);
  }

  async getLinkByUrl(dbCtx, url) {
    const _scope = _fileScope('getLinkByUrl');
    this.logger.debug(_scope, 'called', { url });

    let link = this.statement.getLinkByUrl.get({ url });
    link = SQLiteDatabase._deOphidiate(link);

    this.logger.debug(_scope, 'get', { link });
    return SQLiteDatabase._linkToNative(link);
  }

  async accessLink(dbCtx, id) {
    const _scope = _fileScope('accessLink');
    this.logger.debug(_scope, 'called', { id });

    let link = this.statement.getLinkById.get({ id });
    link = SQLiteDatabase._deOphidiate(link);

    this.logger.debug(_scope, 'get', { id, link });

    if (link) {
      const info = this.statement.incrementLinkAccess.run({ id });
      this.logger.debug(_scope, 'increment', { id, info });
      if (info.changes != 1) {
        this.logger.error(_scope, 'failed to increment link access', { id, link, info });
        throw new DBErrors.UnexpectedResult();
      }
    }
    return SQLiteDatabase._linkToNative(link);
  }

  async expireLink(dbCtx, id, expires) {
    const _scope = _fileScope('expireLink');
    this.logger.debug(_scope, 'called', { id });

    const info = this.statement.expireLink.run({ id, expires });
    if (info.changes != 1) {
      throw new DBErrors.UnexpectedResult();
    }
    return this._sqliteInfo(info);
  }

  async updateLink(dbCtx, id, url) {
    const _scope = _fileScope('updateLink');
    this.logger.debug(_scope, 'called', { id, url });

    const info = this.statement.updateLink.run({ id, url });
    if (info.changes != 1) {
      throw new DBErrors.UnexpectedResult();
    }
    return this._sqliteInfo(info);
  }

  // eslint-disable-next-line no-unused-vars
  async getAllLinks(dbCtx) {
    const _scope = _fileScope('getAllLinks');
    this.logger.debug(_scope, 'called', { });
  
    let links;
    links = this.statement.linkGetAll.all({});
    links = links.map((l) => SQLiteDatabase._deOphidiate(l));
    this.logger.debug(_scope, 'get', { links });
    links = links.map((l) => SQLiteDatabase._linkToNative(l));
    return links;
  }

}

module.exports = SQLiteDatabase;
