'use strict';

const Authenticator = require('./authenticator');
const common = require('./common');
const Enum = require('./enum');
const { ResponseError, ServeStaticFile, SlugGeneratorExhausted } = require('./errors');
const { makeSlugGenerator } = require('./slug');
const Template = require('./template');

const _fileScope = common.fileScope(__filename);

const defaultOptions = {
  pageTitle: require('../package.json').name,
  logoUrl: '/static/muRL.png',
  selfBaseUrl: '',
  staticDirectory: '/static/',
};

class Manager {
  constructor(logger, db, options = {}) {
    common.setOptions(this, defaultOptions, options);
    this.logger = logger;
    this.db = db;
    this.makeSlugGenerator = makeSlugGenerator;
  }


  /**
   * Render root content. Shared by GET and POST.
   * @param {Object} ctx 
   * @param {string} ctx.authToken
   * @param {string} ctx.createdLink
   * @param {string} ctx.message
   * @param {string} ctx.sourceLink
   */
  rootContent(ctx) {
    // eslint-disable-next-line sonarjs/no-small-switch
    switch (ctx.responseType) {
      case Enum.ContentType.ApplicationJson:
        return JSON.stringify({
          link: ctx.createdLink,
          src: ctx.sourceLink,
          token: ctx.authToken,
        });

      case Enum.ContentType.TextHTML:
      default:
        return Template.rootHTML(ctx, this.pageTitle, this.logoUrl);
    }
  }


  /**
   * GET request for root.
   * NB only this manager gets req, to access headers for cache response.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async getRoot(req, res, ctx) {
    const _scope = _fileScope('getRoot');
    this.logger.debug(_scope, 'called', { ctx });

    const content = this.rootContent(ctx);

    /* Since this GET response will never change while server is running,
     * set headers to make it cacheable.
     */
    if (!this.startTime) {
      const now = new Date();
      this.startTime = now.toGMTString();
      this.startTimeMs = now.getTime();
      this.getRootETag = common.generateETag(undefined, undefined, content);
    }
    res.setHeader(Enum.Header.LastModified, this.startTime);
    res.setHeader(Enum.Header.ETag, this.getRootETag);

    if (common.isClientCached(req, this.startTimeMs, this.getRootETag)) {
      this.logger.debug(_scope, 'client cached response', { ctx });
      res.statusCode = 304;
      res.end();
      return;
    }
    this.logger.debug(_scope, 'response', { ctx });
    res.end(content);
  }


  /**
   * The heart of the littler.
   * TODO: other styles of generators
   * @param {*} dbCtx 
   */
  async _getNewIdentifier(dbCtx, url) {
    const idGen = this.makeSlugGenerator(url);
    let id, done;
    do {
      ({ done, value: id } = await idGen.next());
      if (done || !id) {
        throw new SlugGeneratorExhausted();
      }
    } while (await this.db.getLinkById(dbCtx, id));
    return id;
  }


  /**
   * Validates the ctx.sourceLink is acceptable to add.
   * @param {object} ctx 
   * @param {string} ctx.sourceLink
   * @param {string=} ctx.authenticationId
   */
  _validateContextURL(ctx) {
    const _scope = _fileScope('_validateContextURL');

    // Admin is allowed to create locally served entries.
    if (ctx.sourceLink && ctx.sourceLink.startsWith(this.staticDirectory)
    &&  ctx.authenticationId) {
      return;
    }
    // Otherwise, url must be complete.
    try {
      new URL(ctx.sourceLink);
    } catch (e) {
      this.logger.debug(_scope, 'invalid url', { error: e, ctx });
      throw new ResponseError(Enum.ErrorResponse.InvalidURLParameter);
    }
  }


  /**
   * Get request for link, redirect.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async getById(res, ctx) {
    const _scope = _fileScope('getById');
    this.logger.debug(_scope, 'called', { ctx });

    let link;
    await this.db.context(async (dbCtx) => {
      link = await this.db.accessLink(dbCtx, ctx.params.id);
    });
    if (!link) {
      this.logger.info(_scope, 'no such link', { ctx });
      throw new ResponseError(Enum.ErrorResponse.InvalidLinkId);
    }
    this.logger.debug(_scope, 'found link', { link, ctx });

    if (link.expires && link.expires <= Date.now()) {
      this.logger.info(_scope, 'link expired', { link, ctx });
      const expiresDate = new Date(link.expires * 1000).toGMTString();
      res.setHeader(Enum.Header.LastSeen, expiresDate);
      throw new ResponseError(Enum.ErrorResponse.Gone);
    }

    if (link.url.startsWith(this.staticDirectory)) {
      /* This is only a little awkward: throw back up to service to let the
       * dingus handle the static file serving.
       */
      const fileName = link.url.slice(this.staticDirectory.length);
      this.logger.info(_scope, 'static link', { link, ctx });
      throw new ServeStaticFile(fileName);
    }

    this.logger.info(_scope, 'redirect link', { link, ctx });

    res.statusCode = 307; // Temporary Redirect
    res.setHeader('Location', link.url);
    res.end('Redirecting to ' + link.url);
  }


  /**
   * POST request for root, create new link.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async postRoot(res, ctx) {
    const _scope = _fileScope('postRoot');
    this.logger.debug(_scope, 'called', { ctx });

    let content;
    ctx.sourceLink = ctx.parsedBody.url;
    if (!ctx.sourceLink) {
      throw new ResponseError(Enum.ErrorResponse.BadRequest, 'missing required parameter');
    }
    this._validateContextURL(ctx);
    await this.db.context(async (dbCtx) => {
      await this.db.transaction(dbCtx, async (txCtx) => {
        const existingLink = await this.db.getLinkByUrl(txCtx, ctx.sourceLink);
        if (existingLink) {
          res.statusCode = 200; // OK
          ctx.createdLink = `${this.selfBaseUrl}/${existingLink.id}`;
          if (existingLink.expires < Date.now()) {
            await this.db.expireLink(txCtx, existingLink.id, null);
            ctx.authToken = existingLink.authToken;
            this.logger.info(_scope, 'un-expired link', { existingLink, ctx });
          }
        } else {
          res.statusCode = 201; // Created
          const id = await this._getNewIdentifier(txCtx, ctx.sourceLink);
          ctx.authToken = await Authenticator.generateToken();
          await this.db.insertLink(txCtx, id, ctx.sourceLink, ctx.authToken);
          ctx.createdLink = `${this.selfBaseUrl}/${id}`;
          this.logger.info(_scope, 'created new link', { id, ctx });
        }
        content = this.rootContent(ctx);
        return res.end(content);
      });
    });
  }


  /**
   * Put request for link, create with id, or update url.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async putById(res, ctx) {
    const _scope = _fileScope('putById');
    this.logger.debug(_scope, 'called', { ctx });

    ctx.sourceLink = ctx.parsedBody.url;
    if (!ctx.sourceLink) {
      throw new ResponseError(Enum.ErrorResponse.BadRequest, 'missing required parameter');
    }
    this._validateContextURL(ctx);
    await this.db.context(async (dbCtx) => {
      await this.db.transaction(dbCtx, async (txCtx) => {
        const existingLink = await this.db.getLinkById(txCtx, ctx.params.id);
        if (!existingLink) {
          // To create a specific link, need admin authentication.
          if (!ctx.authenticationId) {
            this.logger.debug(_scope, 'cannot put without valid credentials', { ctx });
            throw new ResponseError(Enum.ErrorResponse.Forbidden);
          }

          res.statusCode = 201; // Created
          ctx.authToken = await Authenticator.generateToken();
          this.db.insertLink(txCtx, ctx.params.id, ctx.sourceLink, ctx.authToken);
          ctx.createdLink = `${this.selfBaseUrl}/${ctx.params.id}`;
          this.logger.info(_scope, 'created new link', { ctx });
        } else {
          // Update link
          await this.db.insertLink(txCtx, existingLink.id, ctx.sourceLink, existingLink.token);
          ctx.createdLink = `${this.selfBaseUrl}/${existingLink.id}`;
          this.logger.info(_scope, 'updated link', { ctx });
        }
      });
    });

    const content = this.rootContent(ctx);
    return res.end(content);
  }


  /**
   * Delete request for link, expire.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async deleteById(res, ctx) {
    const _scope = _fileScope('deleteById');
    this.logger.debug(_scope, 'called', { ctx });

    let link;
    await this.db.context(async (dbCtx) => {
      await this.db.transaction(dbCtx, async (txCtx) => {
        link = await this.db.getLinkById(txCtx, ctx.params.id);
        if (!link) {
          this.logger.debug(_scope, 'no such id', { ctx });
          throw new ResponseError(Enum.ErrorResponse.InvalidLinkId);
        }
        if (link.expires && link.expires < Date.now()) {
          res.statusCode = 304; // Not Modified
          return;
        }
        const when = Date.now() / 1000;
        await this.db.expireLink(txCtx, ctx.params.id, when);
        this.logger.info(_scope, 'expired link', { ctx, when });
        res.statusCode = 204; // No Content
      });
    });
    res.end();
  }


  /**
   * Render link information.
   * @param {*} ctx 
   * @param {*} details 
   */
  infoContent(ctx, details) {
    switch (ctx.responseType) {
      case Enum.ContentType.ApplicationJson:
        return JSON.stringify(details);

      case Enum.ContentType.TextHTML:
      default:
        return Template.infoHTML(ctx, details, this.pageTitle);
    }
  }


  /**
   * Get request for link information.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async getByIdInfo(res, ctx) {
    const _scope = _fileScope('getByIdInfo');
    this.logger.debug(_scope, 'called', { ctx });

    let link;
    await this.db.context(async (dbCtx) => {
      link = await this.db.getLinkById(dbCtx, ctx.params.id);
      if (!link) {
        this.logger.debug(_scope, 'no such link', { ctx });
        throw new ResponseError(Enum.ErrorResponse.InvalidLinkId);
      }
      this.logger.info(_scope, 'link found', { ctx, link });
    });

    res.end(this.infoContent(ctx, link));
  }


  /** Render report content.
   * @param {object} ctx
   * @param {object[]} links
   */
  reportContent(ctx, links) {
    switch (ctx.responseType) {
      case Enum.ContentType.ApplicationJson:
        return JSON.stringify(links);

      case Enum.ContentType.TextHTML:
      default:
        return Template.reportHTML(ctx, links, this.pageTitle);
    }
  }


  /**
   * Get request for link information.
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async getAdminReport(res, ctx) {
    const _scope = _fileScope('getAdminReport');
    this.logger.debug(_scope, 'called', { ctx });

    let links;
    await this.db.context(async (dbCtx) => {
      links = await this.db.getAllLinks(dbCtx);
      this.logger.info(_scope, 'found links', { ctx, links });
    });

    res.end(this.reportContent(ctx, links));
  }

}

module.exports = Manager;