'use strict';

/**
 * Here we extend the base API server to define our routes and any route-specific
 * behavior (middlewares) before handing off to the manager.
 */

const { Dingus } = require('@squeep/api-dingus');
const path = require('path');
const common = require('./common');
const Enum = require('./enum');
const { ServeStaticFile } = require('./errors');
const Authenticator = require('./authenticator');
const Manager = require('./manager');

const _fileScope = common.fileScope(__filename);

const defaultOptions = {
  ignoreTrailingSlash: true,
  staticDirectory: path.join(__dirname, '..', 'static'),
  createRequiresAuth: false,
  proxyPrefix: '',
};

class Service extends Dingus {
  constructor(logger, db, options = {}) {
    super(logger, { ...defaultOptions, ...options });
    common.setOptions(this, defaultOptions, options);

    this.authenticator = new Authenticator(logger, db, options.authenticator);
    this._postRootAuth = this.authenticator[this.createRequiresAuth ? 'required' : 'optional'].bind(this.authenticator);

    this.manager = new Manager(logger, db, options.manager);

    this.responseTypes.push(
      // 'other/handled-type',
    );

    this.on('POST', '/', this.handlerPostRoot.bind(this));
    this.on(['GET', 'HEAD'], '/', this.handlerGetRoot.bind(this));
    this.on(['GET', 'HEAD'], '/:id', this.handlerGetId.bind(this));
    this.on('DELETE', '/:id', this.handlerDeleteId.bind(this));
    this.on('PUT', '/:id', this.handlerPutId.bind(this));
    this.on(['GET', 'HEAD'], '/:id/info', this.handlerGetIdInfo.bind(this));
    this.on(['GET', 'HEAD'], '/static/:file', this.handlerGetStatic.bind(this));
    this.on(['GET', 'HEAD'], '/admin/report', this.handlerGetAdminReport.bind(this));
  }


  /**
   * Return a content-type appropriate rendering of an errorResponse object.
   * @param {string} contentType content type of response
   * @param {ResponseError|Exception} err 
   */
  renderError(contentType, err) {
    // eslint-disable-next-line sonarjs/no-small-switch
    switch (contentType) {
      // Handle any additional content types here

      default:
        return super.renderError(contentType, err);
      }
  }


  /**
   * Parse rawBody from ctx as contentType into parsedBody.
   * @param {string} contentType 
   * @param {object} ctx 
   */
  parseBody(contentType, ctx, rawBody) {
    // eslint-disable-next-line sonarjs/no-small-switch
    switch (contentType) {
      // Handle any additional content types here

      default:
        super.parseBody(contentType, ctx, rawBody);
    }
  }


  /**
   * Called before every request handler.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async preHandler(req, res, ctx) {
    super.preHandler(req, res, ctx);
    Dingus.setEndBodyHandler(req, res, ctx, this._endHandler.bind(this));
  }

  /**
   * Do anything needed before sending.
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  _endHandler(req, res, ctx) {
    this.authenticator.signResponse(req, res, ctx);
  }

  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerPostRoot(req, res, ctx) {
    const _scope = _fileScope('handlerPostRoot');
    this.logger.debug(_scope, 'called', { req, ctx });

    this.setResponseType(this.responseTypes, req, res, ctx);
    await this.ingestBody(req, res, ctx, {
      maximumBodySize: 1024 * 8,
    });
    await this._postRootAuth(req, res, ctx);

    await this.manager.postRoot(res, ctx);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerGetRoot(req, res, ctx) {
    const _scope = _fileScope('handlerGetRoot');
    const responseTypes = [
      Enum.ContentType.TextHTML,
    ];
    this.logger.debug(_scope, 'called', { req, ctx });

    Dingus.setHeadHandler(req, res, ctx);

    this.setResponseType(responseTypes, req, res, ctx);
    this.authenticator.optional(req, res, ctx);

    // NB special case for this handler, pass in req so it can check headers
    await this.manager.getRoot(req, res, ctx);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerGetId(req, res, ctx) {
    const _scope = _fileScope('handlerGetId');
    this.logger.debug(_scope, 'called', { req, ctx });
  
    Dingus.setHeadHandler(req, res, ctx);

    this.setResponseType(this.responseTypes, req, res, ctx);
    this.authenticator.optional(req, res, ctx);

    try {
      await this.manager.getById(res, ctx);
    } catch (e) {
      if (e instanceof ServeStaticFile
      &&  e.file) {
        return await this.serveFile(req, res, ctx, this.staticDirectory, e.file);
      }
      throw e;
    }
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerDeleteId(req, res, ctx) {
    const _scope = _fileScope('handlerDeleteId');
    this.logger.debug(_scope, 'called', { req, ctx });
  
    this.setResponseType(this.responseTypes, req, res, ctx);
    await this.authenticator.required(req, res, ctx);

    await this.manager.deleteById(res, ctx);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerPutId(req, res, ctx) {
    const _scope = _fileScope('handlerPutId');
    this.logger.debug(_scope, 'called', { req, ctx });

    this.setResponseType(this.responseTypes, req, res, ctx);
    await this.ingestBody(req, res, ctx);
    await this.authenticator.required(req, res, ctx);

    await this.manager.putById(res, ctx);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerGetIdInfo(req, res, ctx) {
    const _scope = _fileScope('handlerGetIdInfo');
    this.logger.debug(_scope, 'called', { req, ctx });
  
    Dingus.setHeadHandler(req, res, ctx);

    this.setResponseType(this.responseTypes, req, res, ctx);
    this.authenticator.optional(req, res, ctx);

    await this.manager.getByIdInfo(res, ctx);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerGetStatic(req, res, ctx) {
    const _scope = _fileScope('handlerGetStatic');
    this.logger.debug(_scope, 'called', { req, ctx });

    Dingus.setHeadHandler(req, res, ctx);

    // We set default response type to handle any errors, but will override for actual static content type.
    this.setResponseType(this.responseTypes, req, res, ctx);
    this.authenticator.optional(req, res, ctx);

    await this.serveFile(req, res, ctx, this.staticDirectory, ctx.params.file);
  }


  /**
   * @param {http.ClientRequest} req 
   * @param {http.ServerResponse} res 
   * @param {object} ctx 
   */
  async handlerGetAdminReport(req, res, ctx) {
    const _scope = _fileScope('handlerAdminReport');
    this.logger.debug(_scope, 'called', { req, ctx });
  
    Dingus.setHeadHandler(req, res, ctx);

    this.setResponseType(this.responseTypes, req, res, ctx);
    await this.authenticator.required(req, res, ctx);

    await this.manager.getAdminReport(res, ctx);
  }


}

module.exports = Service;
