/* eslint-disable capitalized-comments */
/* eslint-env mocha */

'use strict';

const assert = require('assert');
const sinon = require('sinon'); // eslint-disable-line node/no-unpublished-require

const noExpectedException = 'did not get expected exception';

describe('PostgresDatabase', function () {
  const PostgresDatabase = require('../../../../src/db/postgres');
  let pgpStub, db, logger, options, dbCtx;

  beforeEach(function () {

    pgpStub = () => {
      const stub = {
        result: sinon.stub(),
        all: sinon.stub(),
        get: sinon.stub(),
        run: sinon.stub(),
        one: sinon.stub(),
        manyOrNone: sinon.stub(),
        oneOrNone: sinon.stub(),
        query: sinon.stub(),
        batch: sinon.stub(),
      };
      stub.tx = async (fn) => await fn(stub);
      stub.txIf = async (fn) => await fn(stub);
      stub.task = async (fn) => await fn(stub);
      return stub;
    };
    pgpStub.utils = {
      enumSql: () => ({}),
    };
    pgpStub.QueryFile = class {};
    pgpStub.end = () => {},

    logger = {
      debug: sinon.stub(),
      error: sinon.stub(),
    };
    options = {};
    db = new PostgresDatabase(logger, options, pgpStub);
    db.statement = {};
    dbCtx = undefined;
  });

  describe('context', function () {
    it('covers', async function () {
      const fn = sinon.stub();
      await db.context(fn);
      assert(fn.called);
    });
  }); // context

  describe('transaction', function () {
    it('covers', async function () {
      const fn = sinon.stub();
      await db.transaction(undefined, fn);
      assert(fn.called);
    });
  }); // transaction

  describe('getAuthById', function () {
    let id;
    it('stubbed success', async function () {
      const expected = {
        id: 'id',
        secret: 'secret',
        password: 'password',
       };
      id = 'id';
      db.db.oneOrNone.resolves(expected);
      const result = await db.getAuthById(dbCtx, id);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      db.db.oneOrNone.rejects(expectedExeption);
      try {
        await db.getAuthById(dbCtx, id);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // getAuthById

  describe('_epochFix', function () {
    it('clamps infinity', function () {
      const epoch = Infinity;
      const expected = Number.MAX_SAFE_INTEGER;
      const result = PostgresDatabase._epochFix(epoch);
      assert.strictEqual(result, expected);
    });
    it('clamps negative infinity', function () {
      const epoch = -Infinity;
      const expected = 0;
      const result = PostgresDatabase._epochFix(epoch);
      assert.strictEqual(result, expected);
    });
    it('returns number', function () {
      const epoch = 123;
      const expected = 123;
      const result = PostgresDatabase._epochFix(epoch);
      assert.strictEqual(result, expected);
    });
  }); // _epochFix

  describe('_linkToNative', function () {
    it('handles missing link', function () {
      const link = undefined;
      const expected = undefined;
      const result = PostgresDatabase._linkToNative(link);
      assert.deepStrictEqual(result, expected);
    });
    it('converts epochs', function () {
      const link = {
        id: 'id',
        url: 'url',
        expires: null,
        lastAccess: -Infinity,
        created: 123,
      };
      const expected = {
        id: 'id',
        url: 'url',
        expires: null,
        lastAccess: 0,
        created: 123,
      };
      const result = PostgresDatabase._linkToNative(link);
      assert.deepStrictEqual(result, expected);
    });
  }); // _linkToNative

  describe('insertLink', function () {
    let id, url, authToken;
    it('stubbed success', async function () {
      const returns = {
        rowCount: 0,
        rows: [],
        duration: 0,
      };
      id = 'id';
      db.db.result.resolves(returns);
      const expected = {
        changes: 0,
        duration: 0,
        lastInsertRowid: undefined,
      };
      const result = await db.insertLink(dbCtx, id, url, authToken);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      db.db.result.rejects(expectedExeption);
      try {
        await db.insertLink(dbCtx, id, url, authToken);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // insertLink

  describe('getLinkById', function () {
    let id;
    it('stubbed success', async function () {
      const expected = {
        id: 'id',
        url: 'url',
        created: 0,
        expires: 0,
        lastAccess: 0,
       };
      id = 'id';
      db.db.oneOrNone.resolves(expected);
      const result = await db.getLinkById(dbCtx, id);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      db.db.oneOrNone.rejects(expectedExeption);
      try {
        await db.getLinkById(dbCtx, id);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // getLinkById

  describe('getLinkByUrl', function () {
    let url;
    it('stubbed success', async function () {
      const expected = {
        id: 'id',
        url: 'url',
        created: 0,
        expires: 0,
        lastAccess: 0,
       };
      url = 'url';
      db.db.oneOrNone.resolves(expected);
      const result = await db.getLinkByUrl(dbCtx, url);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      url = 'url';
      db.db.oneOrNone.rejects(expectedExeption);
      try {
        await db.getLinkByUrl(dbCtx, url);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // getLinkByUrl

  describe('accessLink', function () {
    let id;
    it('stubbed success', async function () {
      const expected = {
        id: 'id',
        url: 'url',
        created: 0,
        expires: 0,
        lastAccess: 0,
       };
      id = 'id';
      db.db.oneOrNone.resolves(expected);
      const result = await db.accessLink(dbCtx, id);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      db.db.oneOrNone.rejects(expectedExeption);
      try {
        await db.accessLink(dbCtx, id);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // accessLink

  describe('expireLink', function () {
    let id, expires;
    it('stubbed success', async function () {
      const returns = {
        rowCount: 1,
        rows: [ { id: 1 } ],
        duration: 10,
      };
      const expected = {
        changes: 1,
        duration: 10,
        lastInsertRowid: 1,
       };
      id = 'id';
      expires = null;
      db.db.result.resolves(returns);
      const result = await db.expireLink(dbCtx, id, expires);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      expires = null;
      db.db.result.rejects(expectedExeption);
      try {
        await db.expireLink(dbCtx, id, expires);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // expireLink

  describe('updateLink', function () {
    let id, expires;
    it('stubbed success', async function () {
      const returns = {
        rowCount: 1,
        rows: [ { id: 1 } ],
        duration: 10,
      };
      const expected = {
        changes: 1,
        duration: 10,
        lastInsertRowid: 1,
       };
      id = 'id';
      expires = null;
      db.db.result.resolves(returns);
      const result = await db.updateLink(dbCtx, id, expires);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      id = 'id';
      expires = null;
      db.db.result.rejects(expectedExeption);
      try {
        await db.updateLink(dbCtx, id, expires);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // updateLink

  describe('getAllLinks', function () {
    it('stubbed success', async function () {
      const expected = [
        {
          id: 'id',
          url: 'url',
          created: 0,
          expires: 0,
          lastAccess: 0,
        },
      ];
      db.db.manyOrNone.resolves(expected);
      const result = await db.getAllLinks(dbCtx);
      assert.deepStrictEqual(result, expected);
    });
    it('stubbed failure', async function () {
      const expectedExeption = new Error('blah');
      db.db.manyOrNone.rejects(expectedExeption);
      try {
        await db.getAllLinks(dbCtx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.deepStrictEqual(e, expectedExeption, noExpectedException);
      }
    });
  }); // getAllLinks

});