/* eslint-env mocha */
/* eslint-disable capitalized-comments */

'use strict';

const assert = require('assert');
const sinon = require('sinon'); // eslint-disable-line node/no-unpublished-require

const Manager = require('../../src/manager');
const common = require('../../src/common');
const Enum = require('../../src/enum');
const { ServeStaticFile, SlugGeneratorExhausted } = require('../../src/errors');

const noExpectedException = 'did not get expected exception';

describe('Manager', function () {
  let manager, options;
  let res, ctx;

  beforeEach(function () {
    const stubLogger = {
      debug: sinon.stub(),
      error: sinon.stub(),
      info: sinon.stub(),
    };
    const stubDb = {
      context: async (fn) => await fn({}),
      transaction: async (_dbCtx, fn) => await fn({}),
      getLinkById: sinon.stub(),
      accessLink: sinon.stub(),
      getLinkByUrl: sinon.stub(),
      expireLink: sinon.stub(),
      insertLink: sinon.stub(),
      getAllLinks: sinon.stub(),
    };
    options = {};
    res = {
      end: sinon.stub(),
      setHeader: sinon.stub(),
    };
    ctx = {
      params: {},
    };
    manager = new Manager(stubLogger, stubDb, options);

  });
  afterEach(function () {
    sinon.restore();
  });

  it('instantiates', function () {
    assert(manager);
  });

  it('defaults options', function () {
    manager = new Manager({}, {});
  });

  describe('rootContent', function () {
    it('generates content for empty context', function () {
      const result = manager.rootContent(ctx);
      assert(result.length);
    });
    it('generates json content', function () {
      ctx.responseType = Enum.ContentType.ApplicationJson;
      const result = manager.rootContent(ctx);
      assert(result.length);
      JSON.parse(result);
    });
    it('generates html content', function () {
      ctx.responseType = Enum.ContentType.TextHTML;
      const result = manager.rootContent(ctx);
      assert(result.length);
    });
    it('includes context fields', function () {
      ctx.createdLink = 'http://example.com/foo';
      ctx.authToken = 'token';
      ctx.message = 'message';
      ctx.sourceLink = 'http://source.example.com/';
      const result = manager.rootContent(ctx);
      assert(result.length);
    });
  }); // rootContent

  describe('getRoot', function () {
    let req;
    beforeEach(function () {
      sinon.stub(common, 'isClientCached');
      req = {};
    });
    it('normal response', async function () {
      common.isClientCached.returns(false);
      await manager.getRoot(req, res, ctx);
      assert(res.end.called);
    });
    it('repeat response', async function () {
      manager.startTime = (new Date()).toGMTString();
      common.isClientCached.returns(true);
      await manager.getRoot(req, res, ctx);
      assert(res.end.called);
    });
    it('cached response', async function () {
      common.isClientCached.returns(true);
      await manager.getRoot(req, res, ctx);
      assert(res.end.called);
      assert.strictEqual(res.statusCode, 304);      
    });
  }); // getRoot

  describe('_getNewIdentifier', function () {
    const url = 'http://example.com/bar';
    let dbCtx, nextStub;
    beforeEach(function () {
      dbCtx = {};
      nextStub = sinon.stub();
      manager.db.getLinkById.onCall(0).resolves({ id:'existing' }).onCall(1).resolves();
      sinon.stub(manager, 'makeSlugGenerator').callsFake(() => {
        return {
          next: nextStub,
        };
      });
    });
    it('gets identifiers', async function () {
      nextStub.resolves({ value: 'slug', done: false });
      const result = await manager._getNewIdentifier(dbCtx, url);
      assert(result);
      assert.strictEqual(nextStub.callCount, 2);
    });
    it('handles empty slug', async function () {
      nextStub.resolves({ value: '', done: false });
      try {
        await manager._getNewIdentifier(dbCtx, url);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof SlugGeneratorExhausted, noExpectedException);
      }
    });
    it('handles end of generator', async function () {
      nextStub.resolves({ value: 'blah', done: true });
      try {
        await manager._getNewIdentifier(dbCtx, url);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof SlugGeneratorExhausted, noExpectedException);
      }
    });
  }); // _getNewIdentifier

  describe('_validateContextURL', function () {
    it('allows admin to create local static link', function () {
      ctx.sourceLink = `${manager.staticDirectory}file.txt`;
      ctx.authenticationId = 'awoo';
      manager._validateContextURL(ctx);
    });
    it('accepts valid url', function () {
      ctx.sourceLink = 'http://example.com/file.txt';
      manager._validateContextURL(ctx);
    });
    it('rejects missing url', function () {
      try {
        manager._validateContextURL(ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, Enum.ErrorResponse.InvalidURLParameter.errorMessage, noExpectedException);
      }
    });
    it('rejects invalid url', function () {
      ctx.sourceLink = 'not a url';
      try {
        manager._validateContextURL(ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, Enum.ErrorResponse.InvalidURLParameter.errorMessage, noExpectedException);
      }
    });
  }); // _validateContextURL

  describe('getById', function () {
    let link;
    beforeEach(function () {
      link = undefined;
    });
    it('handles missing link', async function () {
      try {
        await manager.getById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Not Found');
      }
    });
    it('handles expired link', async function () {
      link = {
        expires: 1,
      };
      manager.db.accessLink.resolves(link)
      try {
        await manager.getById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Gone');
      }
    });
    it('handles local static link', async function () {
      const file = 'foop.txt';
      link = {
        url: `${manager.staticDirectory}${file}`,
      };
      manager.db.accessLink.resolves(link);
      try {
        await manager.getById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert(e instanceof ServeStaticFile);
        assert.strictEqual(e.file, file);
      }
    });
    it('redirects a link', async function () {
      link = {
        url: 'http://example.com/awoo',
      };
      manager.db.accessLink.resolves(link);
      await manager.getById(res, ctx);
      assert.strictEqual(res.statusCode, 307);
    });
  }); // getById

  describe('postRoot', function () {
    beforeEach(function () {
      ctx.parsedBody = {};
    });
    it('requires url parameter', async function () {
      try {
        await manager.postRoot(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Bad Request');
      }
    });
    it('creates a link', async function () {
      ctx.parsedBody.url = 'http://example.com/insert';
      await manager.postRoot(res, ctx);
      assert(manager.db.insertLink.called);
      assert(res.end.called);
    });
    it('returns existing link', async function () {
      ctx.parsedBody.url = 'http://example.com/existing';
      const existingLink = {
        id: 'blah',
        sourceLink: ctx.parsedBody.url,
      };
      manager.db.getLinkByUrl.resolves(existingLink);
      await manager.postRoot(res, ctx);
      assert(!manager.db.insertLink.called);
      assert(!manager.db.expireLink.called);
      assert(res.end.called);
    });
    it('restores expired link', async function () {
      ctx.parsedBody.url = 'http://example.com/expired';
      const existingLink = {
        id: 'blah',
        sourceLink: ctx.parsedBody.url,
        expires: 1,
      };
      manager.db.getLinkByUrl.resolves(existingLink);
      await manager.postRoot(res, ctx);
      assert(!manager.db.insertLink.called);
      assert(manager.db.expireLink.called);
      assert(res.end.called);
    });
  }); // postRoot

  describe('putById', function () {
    let url;
    beforeEach(function () {
      url = 'http://example.com/put';
      ctx.parsedBody = {};
    });
    it('requires url parameter', async function () {
      try {
        await manager.putById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Bad Request');
      }
    });
    it('updates existing', async function () {
      ctx.parsedBody.url = url;
      const existingLink = {
        id: 'blah',
      };
      manager.db.getLinkById.resolves(existingLink);
      await manager.putById(res, ctx);
      assert(manager.db.insertLink.called);
      assert(res.end.called);
    });
    it('does not create without admin', async function () {
      ctx.parsedBody.url = url;
      try {
        await manager.putById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Forbidden');
      }
    });
    it('allows admin creation', async function () {
      ctx.parsedBody.url = url;
      ctx.authenticationId = 'blah';
      await manager.putById(res, ctx);
      assert(manager.db.insertLink.called);
      assert.strictEqual(res.statusCode, 201);
      assert(res.end.called);
    });
  }); // putById

  describe('deleteById', function () {
    it('handles missing id', async function () {
      try {
        await manager.deleteById(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Not Found');
      }
    });
    it('expires link', async function () {
      const existingLink = {
        id: 'awoo',
      };
      manager.db.getLinkById.resolves(existingLink);
      await manager.deleteById(res, ctx);
      assert(manager.db.expireLink.called);
      assert.strictEqual(res.statusCode, 204);
    });
    it('ignores expired link', async function () {
      const existingLink = {
        id: 'awoo',
        expires: 123,
      };
      manager.db.getLinkById.resolves(existingLink);
      await manager.deleteById(res, ctx);
      assert(!manager.db.expireLink.called);
      assert.strictEqual(res.statusCode, 304);
    });
  }); // deleteById

  describe('infoContent', function () {
    let details;
    beforeEach(function () {
      details = {
        created: 1604155860,
        lastAccess: 1604155861,
        accesses: 6,
      };
    });
    it('generates info', function () {
      details.expires = 1604155862;
      const result = manager.infoContent(ctx, details);
      assert(result);
    });
    it('generates json info', function () {
      ctx.responseType = Enum.ContentType.ApplicationJson;
      const result = manager.infoContent(ctx, details);
      JSON.parse(result);
    });
    it('generates html info', function () {
      ctx.responseType = Enum.ContentType.TextHTML;
      const result = manager.infoContent(ctx, details);
      assert(result);
    });
  }); // infoContent

  describe('getByIdInfo', function () {
    it('handles missing link', async function () {
      try {
        await manager.getByIdInfo(res, ctx);
        assert.fail(noExpectedException);
      } catch (e) {
        assert.strictEqual(e.message, 'Not Found');
      }
    });
    it('gets link', async function () {
      const existingLink = {
        id: 'blah',
      };
      manager.db.getLinkById.resolves(existingLink);
      await manager.getByIdInfo(res, ctx);
      assert(res.end.called);
    });
  }); // getByIdInfo

  describe('reportContent', function () {
    let links;
    it('generates report', function () {
      links = [
        {
          id: 'awoo',
          created: 1604155860,
          lastAccess: 1604155861,
          url: 'http://example.com/awoo',
        },
      ];
      const result = manager.reportContent(ctx, links);
      assert(result);
    });
    it('generates json report', function () {
      links = [];
      ctx.responseType = Enum.ContentType.ApplicationJson;
      const result = manager.reportContent(ctx, links);
      JSON.parse(result);
    });
    it('generates html report', function () {
      links = [];
      ctx.responseType = Enum.ContentType.TextHTML;
      const result = manager.reportContent(ctx, links);
      assert(result);
    });
  }); // reportContent

  describe('getAdminReport', function () {
    it('does links', async function () {
      const links = [];
      manager.db.getAllLinks.resolves(links);
      await manager.getAdminReport(res, ctx);
      assert(res.end.called);
    });
  }); // getAdminReport

}); // Manager