/* eslint-env mocha */
/* eslint-disable capitalized-comments */

'use strict';

const assert = require('assert');
// eslint-disable-next-line node/no-unpublished-require
const sinon = require('sinon');

const Service = require('../../src/service');
const { ServeStaticFile } = require('../../src/errors');


describe('service', function () {
  let service, logger, options;
  let req, res, ctx;
  const dbStub = {};

  beforeEach(function () {
    options = {};
    logger = {};
    // logger = console;
    service = new Service(logger, dbStub, options);
    sinon.stub(service.manager);
    sinon.stub(service.authenticator);
    sinon.stub(service, 'setResponseType');
    sinon.stub(service, 'serveFile');
    sinon.stub(service, 'ingestBody').resolves();
    req = {
      getHeader: sinon.stub(),
    };
    res = {
      setHeader: sinon.stub(),
      write: sinon.stub(),
      end: sinon.stub(),
    };
    ctx = {};
  });

  afterEach(function () {
    sinon.restore();
  });

  it('instantiates', function () {
    assert(service);
  });

  it('covers no options', function () {
    const r = new Service(logger, dbStub);
    assert(r);
  });

  describe('renderError', function () {
    it('covers default', function () {
      const contentType = 'text/plain';
      const err = {
        statusCode: 418,
        errorMessage: 'I am a teapot',
        details: 'what are you',
      };
      const result = service.renderError(contentType, err);
      assert.strictEqual(result, 'I am a teapot\r\nwhat are you');
    });
  }); // renderError

  describe('parseBody', function () {
    it('covers default', function () {
      const contentType = 'application/json';
      const someData = { foo: 'bar', quux: 3 };
      ctx.rawBody = JSON.stringify(someData);
      service.parseBody(contentType, ctx);
      assert.deepStrictEqual(ctx.parsedBody, someData);
    });
  }); // parseBody

  describe('preHandler', function () {
    it('covers default', async function () {
      await service.preHandler(req, res, ctx);
    });
  }); // preHandler

  describe('_endHandler', function () {
    beforeEach(function () {
      req = {};
      res = {};
      ctx = {};
    });
    it('covers', function() {
      service._endHandler(req, res, ctx);
      assert(service.authenticator.signResponse.called);
    });
  }); // _endHandler

  describe('handlerPostRoot', function () {
    it('covers public mode', async function () {
      sinon.stub(service, '_postRootAuth');

      await service.handlerPostRoot(req, res, ctx);
      assert(service.manager.postRoot.called);
    });
    it('covers private mode', async function () {
      service = new Service(logger, dbStub, { createRequiresAuth: true });
      sinon.stub(service.manager);
      sinon.stub(service.authenticator);
      sinon.stub(service, '_postRootAuth');
      sinon.stub(service, 'setResponseType');
      sinon.stub(service, 'serveFile');
      sinon.stub(service, 'ingestBody').resolves();

      await service.handlerPostRoot(req, res, ctx);
      assert(service.manager.postRoot.called);
    });
  }); // handlerPostRoot

  describe('handlerGetRoot', function () {
    it('covers', async function () {
      await service.handlerGetRoot(req, res, ctx);
      assert(service.manager.getRoot.called);
    });
  }); // handlerGetRoot

  describe('handlerGetId', function () {
    it('covers', async function () {
      await service.handlerGetId(req, res, ctx);
      assert(service.manager.getById.called);
    });
    it('covers static', async function () {
      service.manager.getById.rejects(new ServeStaticFile('this_file.txt'));
      await service.handlerGetId(req, res, ctx);
      assert(service.manager.getById.called);
      assert(service.serveFile.called);
    });
    it('cover errors', async function () {
      const expectedException = 'blah';
      service.manager.getById.rejects(expectedException);
      try {
        await service.handlerGetId(req, res, ctx);
        assert.fail('did not get expected exception');
      } catch (e) {
        assert.strictEqual(e.name, expectedException, 'did not get expected exception');
      }
      assert(service.manager.getById.called);
    });
  }); // handlerGetId

  describe('handlerDeleteId', function () {
    it('covers', async function () {
      await service.handlerDeleteId(req, res, ctx);
      assert(service.manager.deleteById.called);
    });
  }); // handlerDeleteId

  describe('handlerPutId', function () {
    it('covers', async function () {
      await service.handlerPutId(req, res, ctx);
      assert(service.manager.putById.called);
    });
  }); // handlerPutId

  describe('handlerGetIdInfo', function () {
    it('covers', async function () {
      await service.handlerGetIdInfo(req, res, ctx);
      assert(service.manager.getByIdInfo.called);
    });
  }); // handlerGetIdInfo

  describe('handlerGetStatic', function () {
    it('covers', async function () {
      ctx.params = {
        file: '',
      };
      await service.handlerGetStatic(req, res, ctx);
      assert(service.serveFile.called);
    });
  }); // handlerGetStatic

  describe('handlerGetAdminReport', function () {
    it('covers', async function () {
      await service.handlerGetAdminReport(req, res, ctx);
      assert(service.authenticator.required.called);
      assert(service.manager.getAdminReport.called);
    });
  });

});